package sos.connection;

import java.util.Properties;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.StringTokenizer;
import java.util.Vector;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.PreparedStatement;

import java.io.ByteArrayInputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.File;
import java.io.FileReader;
import java.io.Reader;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import sos.util.SOSClassUtil;
import sos.util.SOSLogger;
import sos.util.SOSStandardLogger;
import sos.util.SOSProfiler;
import sos.util.SOSString;
import sos.util.NullBufferedWriter;

/**
 * <p>Title: </p>
 * <p>Description: Basisklasse f�r Datenbankverbindung</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: SOS GmbH</p>
 * @author <a href="mailto:ghassan.beydoun@sos-berlin.com">Ghassan Beydoun</a>
 * @resource ojdbc14.jar sos.util.jar  
 * @version $Id: SOSConnection.java 1298 2005-09-07 14:00:07Z al $
 */

public abstract class SOSConnection {

    protected Statement statement;

    protected ResultSet resultSet;

    protected String[] replacement = {};

    protected static final String NLS_DE = "DE";

    protected static final String NLS_ISO = "ISO";

    /* default-Schreibweise f�r Spaltennamen */
    protected boolean lowerCase = true;
    
    /** JDBC Driver */
    protected String driver;

    /** url des Datenbankservers */
    protected String url;

    /** Datenbankuser */
    protected String dbuser;

    /** password des Datenbankusers */
    protected String dbpassword;

    /** Name der verwendeten Datenbank */
    protected String dbname;

    /** connection-Objekt der aktuellen Datenbankverbindung */
    public Connection connection;

    protected SOSLogger logger = null;

    public static SOSProfiler profiler = null;

    protected boolean fieldNameUpperCase = true;

    protected boolean tableNameUpperCase = true;
    
    /** wird nur intern verwendet bei getStatements */
    private String beginProcedure = "";
    
    /** diese Zeichen werden vor der Suche nach Hochkommaspaare umgewandelt, und spter zurckgesetzt*/
	private final String replaceBackslash 		= "\\\\'";
	private final String replacementBackslash 	= "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx";
	
	private final String replaceDoubleApostrophe 		= "''";
	private final String replacementDoubleApostrophe 	= "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy";



    /**
     * @throws java.lang.Exception
     */
    public SOSConnection() throws Exception {
    }

    /**
     * @param connection
     * @param logger
     * @throws java.lang.Exception
     */
    public SOSConnection(Connection connection, SOSLogger logger)
            throws Exception {
        this.connection = connection;
        this.logger = logger;
    }

    /**
     * @param connection
     * @throws java.lang.Exception
     */
    public SOSConnection(Connection connection) throws Exception {

        this(connection, new SOSStandardLogger(new NullBufferedWriter(
                new OutputStreamWriter(System.out)), SOSStandardLogger.DEBUG9));

    }

    /**
     * Konstruktor wird von der zu implementierenden Klasse verwendet.
     * 
     * @param configFileName
     *            f�r den verwendeten Datenbankserver
     * @param logger
     *            zum Protokollieren
     * @throws java.lang.Exception
     * @see #SOSConnection(String, String, String, String, SOSLogger)
     */
    public SOSConnection(String configFileName, SOSLogger logger)
            throws Exception {

        Properties config = new Properties();
        InputStream in = null;

        try {

            logger.debug6("calling " + SOSClassUtil.getMethodName());
            if (logger == null) throw new Exception("missing logger.");

            File file = new File(configFileName);
            if (!file.exists())
                    throw (new Exception("make sure that the file ["
                            + configFileName + "] does exist!"));

            in = new BufferedInputStream(new FileInputStream(configFileName));
            logger.debug9(".. configFileName: " + configFileName);
            config.load(in);

            if (SOSString.isEmpty(config.getProperty("driver")))
                    throw new Exception("driver missing.");

            this.driver = config.getProperty("driver");
            if (!SOSString.isEmpty(this.driver)) {
                this.driver = this.driver.trim();
            }

            this.url = config.getProperty("url");
            if (!SOSString.isEmpty(this.url)) {
                this.url = this.url.trim();
            }

            this.dbname = config.getProperty("db");
            if (!SOSString.isEmpty(this.dbname)) {
                this.dbname = this.dbname.trim();
            }

            this.dbuser = config.getProperty("user");
            if (!SOSString.isEmpty(this.dbuser)) {
                this.dbuser = this.dbuser.trim();
            }

            this.dbpassword = config.getProperty("password");
            if (!SOSString.isEmpty(this.dbpassword)) {
                this.dbpassword = this.dbpassword.trim();
            }

            this.logger = logger;

        } catch (Exception e) {
            throw e;
        } finally {
            if (in != null) in.close();
        }
    }

    /**
     * Konstruktor wird von der zu implementierenden Klasse verwendet.
     * 
     * @param driver
     *            des verwendeten Datenbankserver
     * @param url
     *            des Datenbankserver
     * @param dbuser
     *            der Datenbank
     * @param dbpassword
     *            Password des dbusers
     * @throws java.lang.Exception
     * @see #SOSConnection(String, SOSLogger)
     */
    public SOSConnection(String driver, String url, String dbuser,
            String dbpassword, SOSLogger logger) throws Exception {

        if (logger == null)
                throw new Exception(SOSClassUtil.getMethodName()
                        + ": missing logger.");

        logger.debug6("calling " + SOSClassUtil.getMethodName());

        if (SOSString.isEmpty(driver))
                throw new Exception(SOSClassUtil.getMethodName()
                        + ": missing database driver.");

        this.driver = driver;
        this.url = url;
        this.dbuser = dbuser;
        this.dbpassword = dbpassword;
        this.logger = logger;
        logger.debug9(".. driver=" + driver + ", url=" + url + ", dbuser="
                + dbuser);
    }

    /**
     * Konstruktor wird von der zu implementierenden Klasse verwendet.
     * 
     * @param configFileName
     *            f�r den verwendeten Datenbankserver
     * @throws java.lang.Exception
     * @see #SOSConnection(String, String, String, String, SOSLogger)
     * @see #SOSConnection(String, String, String, String)
     */
    public SOSConnection(String configFileName) throws Exception {

        Properties config = new Properties();
        InputStream in = null;

        try {

            this.logger = new SOSStandardLogger(new NullBufferedWriter(
                    new OutputStreamWriter(System.out)),
                    SOSStandardLogger.DEBUG9);

            logger.debug6("calling " + SOSClassUtil.getMethodName());

            File file = new File(configFileName);
            if (!file.exists())
                    throw (new Exception("make sure that the file ["
                            + configFileName + "] does exist!"));

            in = new BufferedInputStream(new FileInputStream(configFileName));
            logger.debug9(".. configFileName: " + configFileName);
            config.load(in);

            if (SOSString.isEmpty(config.getProperty("driver")))
                    throw new Exception("driver missing.");

            this.driver = config.getProperty("driver");
            if (!SOSString.isEmpty(this.driver)) {
                this.driver = this.driver.trim();
            }

            this.url = config.getProperty("url");
            if (!SOSString.isEmpty(this.url)) {
                this.url = this.url.trim();
            }

            this.dbname = config.getProperty("db");
            if (!SOSString.isEmpty(this.dbname)) {
                this.dbname = this.dbname.trim();
            }

            this.dbuser = config.getProperty("user");
            if (!SOSString.isEmpty(this.dbuser)) {
                this.dbuser = this.dbuser.trim();
            }

            this.dbpassword = config.getProperty("password");
            if (!SOSString.isEmpty(this.dbpassword)) {
                this.dbpassword = this.dbpassword.trim();
            }

        } catch (Exception e) {
            throw e;
        } finally {
            if (in != null) in.close();
        }
    }

    /**
     * Konstruktor wird von der zu implementierenden Klasse verwendet.
     * 
     * @param driver
     *            des verwendeten Datenbankserver
     * @param url
     *            des Datenbankserver
     * @param dbuser
     *            der Datenbank
     * @param dbpassword
     *            Password des dbusers
     * @throws java.lang.Exception
     * @see #SOSConnection(String, SOSLogger)
     * @see #SOSConnection(String)
     */
    public SOSConnection(String driver, String url, String dbuser,
            String dbpassword) throws Exception {

        this.logger = new SOSStandardLogger(new NullBufferedWriter(
                new OutputStreamWriter(System.out)), SOSStandardLogger.DEBUG9);

        logger.debug6("calling " + SOSClassUtil.getMethodName());

        if (SOSString.isEmpty(driver))
                throw new Exception(SOSClassUtil.getMethodName()
                        + ": missing database driver.");

        this.driver = driver;
        this.url = url;
        this.dbuser = dbuser;
        this.dbpassword = dbpassword;

        logger.debug9(".. driver=" + driver + ", url=" + url + ", dbuser="
                + dbuser);
    }

    /**
     * Liefert den Namen der Connection-Klasse aus der Konfigurationsdatei
     * 
     * @param configFileName
     *            f�r den verwendeten Datenbankserver
     * @param logger
     *            zum Protokollieren
     * @throws java.lang.Exception
     * @see #SOSConnection(String, String, String, String, SOSLogger)
     */
    public static String getClassName(String configFileName, SOSLogger logger)
            throws Exception {

        Properties config = new Properties();
        InputStream in = null;
        String className = null;

        try {

            if (logger != null) {
                logger.debug9("calling " + SOSClassUtil.getMethodName());
            }
            // if (logger == null) throw new Exception("missing logger.");

            File file = new File(configFileName);
            if (!file.exists())
                    throw (new Exception("configuration file not found: "
                            + configFileName));

            in = new BufferedInputStream(new FileInputStream(configFileName));
            if (logger != null) {
                logger.debug9(".. configuration file: " + configFileName);
            }
            config.load(in);

            if (SOSString.isEmpty(config.getProperty("class")))
                    throw new Exception("class missing.");

            className = config.getProperty("class").trim();

        } catch (Exception e) {
            throw e;
        } finally {
            if (in != null) try {
                in.close();
            } catch (Exception ignore) {
            }
        }
        return className;
    }

    /**
     * liefert eine neue Instanz einer Connection-Klasse der aktuellen
     * Datenbankverbindung zur�ck.
     * 
     * @return connection der aktuellen Datenbankverbindung.
     * @exception falls
     *                ein Datenbankfehler vorliegt.
     */
    public static SOSConnection createInstance(String configFileName,
            SOSLogger logger) throws Exception {

        logger.debug6("calling " + SOSClassUtil.getMethodName());
        String className = SOSConnection.getClassName(configFileName, logger);
        logger.debug9(".. creating instance for: " + className);

        if (className.equalsIgnoreCase("SOSMSSQLConnection")) {
            return new SOSMSSQLConnection(configFileName, logger);
        } else if (className.equalsIgnoreCase("SOSMySQLConnection")) {
            return new SOSMySQLConnection(configFileName, logger);
        } else if (className.equalsIgnoreCase("SOSOracleConnection")) {
            return new SOSOracleConnection(configFileName, logger);
        } else if (className.equalsIgnoreCase("SOSODBCConnection")) {
            return new SOSODBCConnection(configFileName, logger);
        } else {
            throw new Exception(
                    SOSClassUtil.getMethodName()
                            + ": sorry, no object can be derived from unknown connection class: "
                            + className);
        }
    }

    /**
     * liefert eine neue Instanz einer Connection-Klasse der aktuellen
     * Datenbankverbindung zur�ck.
     * 
     * @param className
     * @param connection
     * @param logger
     * @return connection der aktuellen Datenbankverbindung.
     * @throws java.lang.Exception
     */
    public static SOSConnection createInstance(String className,
            Connection connection, SOSLogger logger) throws Exception {

        logger.debug6("calling " + SOSClassUtil.getMethodName());

        if (className.equalsIgnoreCase("SOSMSSQLConnection")) {
            return new SOSMSSQLConnection(connection, logger);
        } else if (className.equalsIgnoreCase("SOSMySQLConnection")) {
            return new SOSMySQLConnection(connection, logger);
        } else if (className.equalsIgnoreCase("SOSOracleConnection")) {
            return new SOSOracleConnection(connection, logger);
        } else if (className.equalsIgnoreCase("SOSODBCConnection")) {
            return new SOSODBCConnection(connection, logger);
        } else {
            throw new Exception(
                    SOSClassUtil.getMethodName()
                            + ": sorry, no object can be derived from unknown connection class: "
                            + className);
        }

    }

    /**
     * liefert eine neue Instanz einer Connection-Klasse der aktuellen
     * Datenbankverbindung zur�ck.
     * 
     * @param className
     * @param connection
     * @return connection der aktuellen Datenbankverbindung.
     * @throws java.lang.Exception
     */
    public static SOSConnection createInstance(String className,
            Connection connection) throws Exception {

        if (className.equalsIgnoreCase("SOSMSSQLConnection")) {
            return new SOSMSSQLConnection(connection);
        } else if (className.equalsIgnoreCase("SOSMySQLConnection")) {
            return new SOSMySQLConnection(connection);
        } else if (className.equalsIgnoreCase("SOSOracleConnection")) {
            return new SOSOracleConnection(connection);
        } else if (className.equalsIgnoreCase("SOSODBCConnection")) {
            return new SOSODBCConnection(connection);
        } else {
            throw new Exception(
                    SOSClassUtil.getMethodName()
                            + ": sorry, no object can be derived from unknown connection class: "
                            + className);
        }

    }

    /**
     * liefert eine neue Instanz einer Connection-Klasse der aktuellen
     * Datenbankverbindung zur�ck.
     * 
     * @return connection der aktuellen Datenbankverbindung.
     * @exception falls
     *                ein Datenbankfehler vorliegt.
     */
    public static SOSConnection createInstance(String configFileName)
            throws Exception {

        String className = SOSConnection.getClassName(configFileName, null);

        if (className.equalsIgnoreCase("SOSMSSQLConnection")) {
            return new SOSMSSQLConnection(configFileName);
        } else if (className.equalsIgnoreCase("SOSMySQLConnection")) {
            return new SOSMySQLConnection(configFileName);
        } else if (className.equalsIgnoreCase("SOSOracleConnection")) {
            return new SOSOracleConnection(configFileName);
        } else if (className.equalsIgnoreCase("SOSODBCConnection")) {
            return new SOSODBCConnection(configFileName);
        } else {
            throw new Exception(
                    SOSClassUtil.getMethodName()
                            + ": sorry, no object can be derived from unknown connection class: "
                            + className);
        }
    }

    /**
     * liefert eine neue Instanz einer Connection-Klasse der aktuellen
     * Datenbankverbindung zur�ck.
     * 
     * @param className
     * @param driver
     * @param url
     * @param dbuser
     * @param dbpassword
     * @return connection der aktuellen Datenbankverbindung.
     * @throws java.lang.Exception
     */
    public static SOSConnection createInstance(String className, String driver,
            String url, String dbuser, String dbpassword) throws Exception {

        if (className.equalsIgnoreCase("SOSMSSQLConnection")) {
            return new SOSMSSQLConnection(driver, url, dbuser, dbpassword);
        } else if (className.equalsIgnoreCase("SOSMySQLConnection")) {
            return new SOSMySQLConnection(driver, url, dbuser, dbpassword);
        } else if (className.equalsIgnoreCase("SOSOracleConnection")) {
            return new SOSOracleConnection(driver, url, dbuser, dbpassword);
        } else if (className.equalsIgnoreCase("SOSODBCConnection")) {
            return new SOSODBCConnection(driver, url, dbuser, dbpassword);
        } else {
            throw new Exception(
                    SOSClassUtil.getMethodName()
                            + ": sorry, no object can be derived from unknown connection class: "
                            + className);
        }

    }

    /**
     * liefert eine neue Instanz einer Connection-Klasse der aktuellen
     * Datenbankverbindung zur�ck.
     * 
     * @param className
     * @param driver
     * @param url
     * @param dbuser
     * @param dbpassword
     * @return connection der aktuellen Datenbankverbindung.
     * @throws java.lang.Exception
     */
    public static SOSConnection createInstance(String className, String driver,
            String url, String dbuser, String dbpassword, SOSLogger logger)
            throws Exception {

        if (className.equalsIgnoreCase("SOSMSSQLConnection")) {
            return new SOSMSSQLConnection(driver, url, dbuser, dbpassword,
                    logger);
        } else if (className.equalsIgnoreCase("SOSMySQLConnection")) {
            return new SOSMySQLConnection(driver, url, dbuser, dbpassword,
                    logger);
        } else if (className.equalsIgnoreCase("SOSOracleConnection")) {
            return new SOSOracleConnection(driver, url, dbuser, dbpassword,
                    logger);
        } else if (className.equalsIgnoreCase("SOSODBCConnection")) {
            return new SOSODBCConnection(driver, url, dbuser, dbpassword,
                    logger);
        } else {
            throw new Exception(
                    SOSClassUtil.getMethodName()
                            + ": sorry, no object can be derived from unknown connection class: "
                            + className);
        }

    }

    /**
     * liefert das connection-Objekt der aktuellen Datenbankverbindung zur�ck.
     * 
     * @return connection der aktuellen Datenbankverbindung.
     * @exception falls
     *                ein Datenbankfehler vorliegt.
     */
    public Connection getConnection() throws Exception {

        if (connection == null)
                throw new Exception(
                        SOSClassUtil.getMethodName()
                                + ": sorry, there is no successful connection established.");

        return this.connection;
    }

    /**
     * liefert einen Satz als HashMap zur�ck.
     * 
     * @param query
     * @return HashMap mit Spaltennamen in Kleinbuchstaben.
     * @throws java.lang.Exception
     * @see #getSingleAsProperties
     */
    public HashMap getSingle(String query) throws Exception {

        HashMap results = new LinkedHashMap();
        Statement stmt = null;
        ResultSet rs = null;
        ResultSetMetaData meta = null;
        int columnCount = 0;
        String key, value;

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            try {
                if (profiler != null) profiler.start(query);
            } catch (Exception e) {
            }

            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            meta = rs.getMetaData();
            columnCount = meta.getColumnCount();

            if (rs.next()) {
                for (int i = 1; i <= columnCount; i++) {
                    key = normalizeKey(meta.getColumnName(i));
                    value = rs.getString(i);
                    if (SOSString.isEmpty(value)) value = "";
                    results.put(normalizeKey(key), value.trim());
                }
            }
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName()
                    + ": query failed: " + query + ": " + e.toString(), e);
        } finally {
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (rs != null) try {
                rs.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return results;
    }

    /**
     * liefert einen Satz als Property zur�ck.
     * 
     * @param query
     * @return Property mit Spaltennamen in Kleinbuchstaben.
     * @throws java.lang.Exception
     * @see #getSingle( String )
     */
    public Properties getSingleAsProperties(String query) throws Exception {

        Properties results = new Properties();
        Statement stmt = null;
        ResultSet rs = null;
        ResultSetMetaData meta = null;
        int columnCount = 0;
        String key, value;

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            if (profiler != null) try {
                profiler.start(query);
            } catch (Exception e) {
            }

            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            meta = rs.getMetaData();
            columnCount = meta.getColumnCount();

            if (rs.next()) {
                for (int i = 1; i <= columnCount; i++) {
                    key = normalizeKey(meta.getColumnName(i));
                    value = rs.getString(i);
                    if (SOSString.isEmpty(value)) value = "";
                    logger.debug9(SOSClassUtil.getMethodName() + ", key= "
                            + key + ", value= " + value);
                    results.setProperty(normalizeKey(key), value.trim());
                }
            }

            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName()
                    + ": query failed: " + query + ": " + e.toString(), e);
        } finally {
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (rs != null) try {
                rs.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return results;
    }

    /**
     * liefert mehrere S�tze als Properties zur�ck.
     * 
     * @param query
     * @return Properties mit Spaltennamen.
     * @throws java.lang.Exception
     * @see #getSingle( String )
     */
    public Properties getArrayAsProperties(String query) throws Exception {

        Properties results = new Properties();
        Statement stmt = null;
        ResultSet rs = null;
        ResultSetMetaData meta = null;
        int columnCount = 0;
        String key, value;

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);
            if (profiler != null) try {
                profiler.start(query);
            } catch (Exception e) {
            }
            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            meta = rs.getMetaData();
            columnCount = meta.getColumnCount();
            while (rs.next()) {

                if (columnCount > 1) {
                    key = rs.getString(1);
                    value = rs.getString(2);
                    if (SOSString.isEmpty(value)) value = "";
                    logger.debug9(SOSClassUtil.getMethodName() + ", key= "
                            + key + ", value= " + value);
                    results.setProperty(normalizeKey(key), value.trim());
                }
            }
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName()
                    + ": query failed: " + query + ": " + e.toString(), e);
        } finally {
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (rs != null) try {
                rs.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return results;
    }

    /**
     * liefert einen scalaren Wert zur�ck. Das Statemnet darf nur ein Feld aus
     * genau einem Satz lesen.
     * 
     * @param query
     * @return String Ergebnis der Abfrage
     * @throws java.lang.Exception
     */
    public String getSingleValue(String query) throws Exception {

        Statement stmt = null;
        ResultSet rs = null;
        ResultSetMetaData meta = null;
        String result = "";
        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());
            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            if (profiler != null) profiler.start(query);

            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            if (rs.next()) {
                result = rs.getString(1);
                if (SOSString.isEmpty(result)) result = "";
            }
            logger.debug9(SOSClassUtil.getMethodName()
                    + ".. successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName()
                    + ": query failed: " + query + ": " + e.toString(), e);
        } finally {
            if (rs != null) try {
                rs.close();
            } catch (Exception e) {
            }
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return result.trim();
    }

    /**
     * liefert die Treffer als ArrayList aus HashMaps zur�ck. Feldnamen sind
     * in der HashMap in Kleinbuchstaben abgespeichert.
     * 
     * @param query
     * @return ArrayList aus HashMap
     * @throws java.lang.Exception
     */
    public ArrayList getArray(String query) throws Exception {

        ArrayList results = new ArrayList();
        HashMap record = new LinkedHashMap();
        Statement stmt = null;
        ResultSet rs = null;
        ResultSetMetaData meta = null;
        String key, value;
        int columnCount = 0;

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName() + ": "
                    + query);

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            if (profiler != null) try {
                profiler.start(query);
            } catch (Exception e) {
            }
            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            meta = rs.getMetaData();
            columnCount = meta.getColumnCount();

            while (rs.next()) {
                for (int i = 1; i <= columnCount; i++) {
                    key = normalizeKey(meta.getColumnName(i));
                    value = rs.getString(meta.getColumnName(i));
                    if (SOSString.isEmpty(value)) value = "";
                    record.put(normalizeKey(key), value.trim());
                }
                results.add(record);
                record = new HashMap();
            } // while

            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName()
                    + ": query failed: " + query + ": " + e.toString(), e);
        } finally {
            if (rs != null) try {
                rs.close();
            } catch (Exception e) {
            }
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return results;
    }

    /**
     * liefert die Treffer als Vector aus HashMaps zur�ck. Feldnamen sind in
     * der HashMap in Kleinbuchstaben abgespeichert.
     * 
     * @param query
     * @return Vector aus HashMap
     * @throws java.lang.Exception
     */
    public Vector getArrayAsVector(String query) throws Exception {

        Vector results = new Vector();
        HashMap record = new LinkedHashMap();
        Statement stmt = null;
        ResultSet rs = null;
        ResultSetMetaData meta = null;
        String key, value;
        int columnCount = 0;

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName() + ": "
                    + query);

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            if (profiler != null) try {
                profiler.start(query);
            } catch (Exception e) {
            }
            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            meta = rs.getMetaData();
            columnCount = meta.getColumnCount();

            while (rs.next()) {
                for (int i = 1; i <= columnCount; i++) {
                    key = normalizeKey(meta.getColumnName(i));
                    value = rs.getString(meta.getColumnName(i));
                    if (SOSString.isEmpty(value)) value = "";
                    record.put(normalizeKey(key), value.trim());
                }
                results.add(record);
                record = new HashMap();
            } // while

            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName()
                    + ": query failed: " + query + ": " + e.toString(), e);
        } finally {
            if (rs != null) try {
                rs.close();
            } catch (Exception e) {
            }
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return results;
    }

    /**
     * Liefert die Trefferliste als ArrayList-Objekt zur�ck. Das Statement
     * darf nur ein Feld aus jeweils einem Satz lesen.
     * 
     * @param query
     * @return ArrayList
     * @throws java.lang.Exception
     * @see #getArray( String )
     */
    public ArrayList getArrayValue(String query) throws Exception {

        ArrayList results = new ArrayList();
        Statement stmt = null;
        ResultSet rs = null;
        String value;
        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);
            if (profiler != null) profiler.start(query);
            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            while (rs.next()) {
                value = rs.getString(1);
                if (!SOSString.isEmpty(value)) results.add(value.trim());
            } // while
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName()
                    + ": query failed: " + query + ": " + e.toString(), e);
        } finally {
            if (rs != null) try {
                rs.close();
            } catch (Exception e) {
            }
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return results;
    }

    /**
     * Aktualisiert ein Blob-Feld
     * 
     * @param tableName
     *            Zieltabelle
     * @param columnName
     *            Zielspalte
     * @param data
     *            Blob-Daten als byte-array
     * @param condition
     *            Bedingung(en) der SQL-Statement
     * @return Anzahl betroffener S�tze
     * @throws java.lang.Exception
     * @see #getBlob( String )
     * @see #getBlob( String, String )
     */
    public long updateBlob(String tableName, String columnName, byte[] data,
            String condition) throws Exception {

        ByteArrayInputStream in = null;
        PreparedStatement pstmt = null;
        StringBuffer query = null;
        String theQuery = null;
        try {

            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            if (tableName == null)
                    throw new NullPointerException("tableName is null.");
            if (columnName == null)
                    throw new NullPointerException("columnName is null.");
            if (data == null || data.length <= 0)
                    throw new NullPointerException("missing data.");

            in = new ByteArrayInputStream(data);

            query = new StringBuffer("UPDATE ");

            if (tableNameUpperCase)
                query.append(tableName.toUpperCase());
            else
                query.append(tableName);

            if (fieldNameUpperCase) {
                query.append(" SET \"");
                query.append(columnName.toUpperCase());
                query.append("\" = ? ");
            } else {
                query.append(" SET ");
                query.append(columnName);
                query.append(" = ? ");
            }

            if (!SOSString.isEmpty(condition)) {
                query.append(" WHERE ");
                query.append(condition);
            }

            theQuery = this.normalizeStatement(query.toString(), replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + theQuery);

            if (profiler != null) try {
                profiler.start(theQuery);
            } catch (Exception e) {
            }

            pstmt = connection.prepareStatement(theQuery);
            pstmt.setBinaryStream(1, in, data.length);
            pstmt.executeUpdate();
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            if (in != null) try {
                in.close();
            } catch (Exception e) {
            }
            if (pstmt != null) try {
                pstmt.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return data.length;
    }

    /**
     * Aktualisiert ein Clob-Feld
     * 
     * @param tableName
     *            Zieltabelle
     * @param columnName
     *            Zielspalte
     * @param data
     *            Clob-Daten
     * @param condition
     *            Bedingung(en) der SQL-Statement
     * @return Anzahl geschriebener Bytes
     * @throws java.lang.Exception
     * @see #getClob( String )
     * @see #getClob( String, String )
     */
    public long updateClob(String tableName, String columnName, String data,
            String condition) throws Exception {

        PreparedStatement pstmt = null;
        StringBuffer query = null;
        long totalBytesWritten = 0;
        String theQuery = null;

        try {

            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            if (SOSString.isEmpty(tableName))
                    throw new NullPointerException("tableName is null.");
            if (SOSString.isEmpty(columnName))
                    throw new NullPointerException("columnName is null.");
            if (SOSString.isEmpty(data))
                    throw new NullPointerException("missing data.");

            query = new StringBuffer("UPDATE ");
            if (tableNameUpperCase)
                query.append(tableName.toUpperCase());
            else
                query.append(tableName);

            if (fieldNameUpperCase) {
                query.append(" SET \"");
                query.append(columnName.toUpperCase());
                query.append("\" = ? ");
            } else {
                query.append(" SET ");
                query.append(columnName);
                query.append(" = ? ");
            }

            if (!SOSString.isEmpty(condition)) {
                query.append(" WHERE ");
                query.append(condition);
            }
            theQuery = this.normalizeStatement(query.toString(), replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + theQuery);

            if (profiler != null) try {
                profiler.start(theQuery);
            } catch (Exception e) {
            }

            pstmt = connection.prepareStatement(theQuery);
            totalBytesWritten = data.length();
            pstmt.setCharacterStream(1, new java.io.StringReader(data),
                    (int) totalBytesWritten);
            pstmt.executeUpdate();
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) profiler.stop("ERROR", e.getMessage());
            throw new Exception(SOSClassUtil.getMethodName() + ": "
                    + e.toString(), e);
        } finally {
            if (pstmt != null) try {
                pstmt.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return totalBytesWritten;
    }

    /**
     * Liefert den Inhalt eines Clob-Felds bei Erfolg zur�ck, sonst leere
     * Zeichenkette.
     * 
     * @param query
     *            die SQL-Abfrage
     * @return den Inhalt des Clob-Felds zur�ck
     * @throws java.lang.Exception
     * @see #updateClob( String, String, String, String )
     * @see #updateClob( String, String, File, String )
     * @see #getClob( String, String )
     */

    public String getClob(String query) throws Exception {

        ResultSet rs = null;
        Statement stmt = null;
        Reader in = null;
        int bytesRead;
        StringBuffer sb = new StringBuffer();
        ;

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());
            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);
            if (profiler != null) try {
                profiler.start(query);
            } catch (Exception e) {
            }
            stmt = connection.createStatement();
            try {
                rs = stmt.executeQuery(query);
            } catch (Exception e) {
                logger
                        .debug6(".. query failed: " + query + ": "
                                + e.toString());
                throw new Exception(SOSClassUtil.getMethodName() + ": "
                        + e.toString());
            }

            if (rs.next()) {
                in = rs.getCharacterStream(1);
                if (in == null) {
                    logger.debug6(".. ResultSet returns NULL value.");
                    return "";
                }

                if ((bytesRead = in.read()) != -1) { // hat CLOB-Feld 0 bytes?
                    sb.append((char) bytesRead);
                } else {
                    logger.debug6(".. CLOB column has 0 bytes.");
                    return "";
                }

                while ((bytesRead = in.read()) != -1) {
                    sb.append((char) bytesRead);
                }
            }
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (in != null) try {
                in.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return sb.toString();
    }

    /**
     * Schreibt den Inhalt eines Clob-Felds in eine Datei.
     * 
     * @param query
     *            die SQL-Abfrage
     * @return Anzahl geschriebener Bytes
     * @throws java.lang.Exception
     * @see #updateClob( String, String, String, String )
     * @see #updateClob( String, String, File, String )
     * @see #getClob( String )
     */
    public long getClob(String query, String fileName) throws Exception {

        FileOutputStream out = null;
        ResultSet rs = null;
        Statement stmt = null;
        int bytesRead = 0;
        long readBytes = 0;
        Reader in = null;
        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            if (profiler != null) profiler.start(query);

            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            if (rs.next()) {

                in = rs.getCharacterStream(1);

                if (in == null) {
                    logger.debug9(".. ResultSet returns NULL value.");
                    return readBytes;
                }

                if ((bytesRead = in.read()) != -1) { // hat CLOB-Feld 0 bytes?
                    out = new FileOutputStream(fileName);
                    out.write(bytesRead);
                    readBytes++;
                } else {
                    logger.debug9(".. CLOB column has 0 bytes.");
                    return readBytes;
                }

                while ((bytesRead = in.read()) != -1) {
                    out.write(bytesRead);
                    readBytes++;
                }
            }
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) profiler.stop("ERROR", e.getMessage());
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            try {
                out.flush();
            } catch (Exception e) {
            }
            if (out != null) try {
                out.close();
            } catch (Exception e) {
            }
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (in != null) try {
                in.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return readBytes;
    }

    /**
     * Aktualisiert ein Clob-Feld aus einer Eingabedatei.
     * 
     * @param tableName
     *            Zieltabelle
     * @param columnName
     *            Zielspalte
     * @param file
     *            Quelle der Clob-Daten
     * @param condition
     *            Bedingung(en) der SQL-Statement
     * @return Anzahl geschriebener Bytes
     * @throws java.lang.Exception
     * @see #updateClob( String, String, String, String )
     * @see #getClob( String, String )
     * @see #getClob( String )
     */
    public long updateClob(String tableName, String columnName, File file,
            String condition) throws Exception {

        PreparedStatement pstmt = null;
        StringBuffer query = null;
        long totalBytesWritten = 0;
        String theQuery = null;
        Reader in = null;

        try {

            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            if (tableName == null)
                    throw new NullPointerException("tableName is null.");
            if (columnName == null)
                    throw new NullPointerException("columnName is null.");
            if (!file.exists()) throw new Exception("file doesn't exist.");

            query = new StringBuffer("UPDATE ");
            if (tableNameUpperCase) {
                query.append(tableName.toUpperCase());
                query.append(" SET \"");
            } else {
                query.append(tableName);
                query.append(" SET ");
            }
            if (fieldNameUpperCase) {
                query.append(columnName.toUpperCase());
                query.append("\" = ? ");
            } else {
                query.append(columnName);
                query.append(" = ? ");
            }
            if (condition != null) {
                query.append(" WHERE ");
                query.append(condition);
            }

            theQuery = this.normalizeStatement(query.toString(), replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + theQuery);

            if (profiler != null) try {
                profiler.start(theQuery);
            } catch (Exception e) {
            }

            pstmt = connection.prepareStatement(theQuery);
            totalBytesWritten = file.length();
            in = new FileReader(file);
            // pstmt.setCharacterStream(1, new FileReader(file), (int)
            // totalBytesWritten);
            pstmt.setCharacterStream(1, in, (int) totalBytesWritten);
            pstmt.executeUpdate();
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            if (pstmt != null) try {
                pstmt.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
            if (in != null) try {
                in.close();
            } catch (Exception e) {
            }

        }
        return totalBytesWritten;
    }

    /**
     * Aktualisiert ein Blob-Feld aus einer Eingabedatei.
     * 
     * @param tableName
     *            Zieltabelle
     * @param columnName
     *            Zielspalte
     * @param fileName
     *            Quelle der Clob-Daten
     * @param condition
     *            Bedingung(en) der SQL-Statement
     * @return Anzahl betroffener S�tze
     * @throws java.lang.Exception
     * @see #updateClob( String, String, String, String )
     * @see #getClob( String, String )
     * @see #getClob( String )
     */
    public long updateBlob(String tableName, String columnName,
            String fileName, String condition) throws Exception {

        PreparedStatement pstmt = null;
        long totalBytesRead = 0;
        InputStream in = null;
        StringBuffer query = null;
        String theQuery = null;

        try {

            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            if (SOSString.isEmpty(tableName))
                    throw new NullPointerException("tableName is null.");
            if (SOSString.isEmpty(columnName))
                    throw new NullPointerException("columnName is null.");
            if (SOSString.isEmpty(fileName))
                    throw new Exception("fileName is null.");

            File file = new File(fileName);
            if (!file.exists())
                    throw new Exception(SOSClassUtil.getMethodName()
                            + ": file doesn't exist.");

            in = new FileInputStream(file);
            query = new StringBuffer("UPDATE ");

            if (tableNameUpperCase)
                query.append(tableName.toUpperCase());
            else
                query.append(tableName);

            if (fieldNameUpperCase) {
                query.append(" SET \"");
                query.append(columnName.toUpperCase());
                query.append("\" = ? ");
            } else {
                query.append(" SET ");
                query.append(columnName);
                query.append(" = ? ");
            }

            if (!SOSString.isEmpty(condition)) {
                query.append(" WHERE ");
                query.append(condition);
            }

            theQuery = normalizeStatement(query.toString(), replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + theQuery);

            if (profiler != null) profiler.start(theQuery);

            pstmt = connection.prepareStatement(theQuery);
            totalBytesRead = file.length();
            logger.debug9(".. length: " + totalBytesRead);
            pstmt.setBinaryStream(1, in, (int) totalBytesRead);
            pstmt.executeUpdate();
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            if (connection != null) connection.rollback();
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            if (pstmt != null) try {
                pstmt.close();
            } catch (Exception e) {
            }
            if (in != null) try {
                in.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return totalBytesRead;

    }

    /**
     * Schreibt den Inhalt eines Blob-Felds in eine Datei.
     * 
     * @param query
     *            die SQL-Abfrage
     * @return Anzahl geschriebener Bytes
     * @throws java.lang.Exception
     * @see #updateBlob( String, String, byte[], String)
     * @see #updateBlob( String, String, String, String)
     * @see #getBlob( String )
     */
    public long getBlob(String query, String fileName) throws Exception {

        InputStream in = null;
        FileOutputStream out = null;
        ResultSet rs = null;
        Statement stmt = null;
        long readBytes = 0;
        int len = 0;

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            if (profiler != null) try {
                profiler.start(query);
            } catch (Exception e) {
            }

            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            if (rs.next()) {
                in = rs.getBinaryStream(1);
                if (in == null) { return readBytes; }

                byte[] buff = new byte[1024];
                if ((len = in.read(buff)) > 0) { // hat BLOB-Feld 0 byte??
                    out = new FileOutputStream(fileName);
                    out.write(buff, 0, len);
                    readBytes += len;
                } else {
                    logger.debug9(".. BLOB column has 0 bytes.");
                    return readBytes;
                }

                while (0 < (len = in.read(buff))) {
                    out.write(buff, 0, len);
                    readBytes += len;
                }
            }
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            if (out != null) try {
                out.close();
            } catch (Exception e) {
            }
            if (in != null) try {
                in.close();
            } catch (Exception e) {
            }
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return readBytes;
    }

    /**
     * Liefert den Inhalt eines BLOB-Felds als byte-array zur�ck.
     * 
     * @param query
     *            die SQL-Abfrage
     * @return Anzahl gelesener Bytes
     * @throws java.lang.Exception
     * @see #getBlob( String, String )
     * @see #updateBlob( String, String, byte[], String )
     * @see #updateBlob( String, String, String, String )
     */
    public byte[] getBlob(String query) throws Exception {

        ResultSet rs = null;
        Statement stmt = null;
        byte[] result = {};
        int bytesRead;

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            if (profiler != null) try {
                profiler.start(query);
            } catch (Exception e) {
            }

            stmt = connection.createStatement();
            rs = stmt.executeQuery(query);
            if (rs.next()) result = rs.getBytes(1);
            if (result == null) {
                logger.debug9(".. BLOB column has 0 bytes.");
                return result;
            }
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            if (stmt != null) try {
                stmt.close();
            } catch (Exception e) {
            }
            if (rs != null) try {
                rs.close();
            } catch (Exception e) {
            }
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        }
        return result;
    }

    /**
     * F�hrt ein Statement aus und beh�lt das ResultSet f�r Aufrufe mit
     * get().
     * 
     * @param query
     * @throws java.lang.Exception
     * @see #execute( String )
     * @see #get()
     */
    public void executeQuery(String query) throws Exception {

        try {

            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);

            if (profiler != null) try {
                profiler.start(query);
            } catch (Exception e) {
            }
            if (this.statement != null) this.statement.close();
            this.statement = connection.createStatement();
            this.resultSet = this.statement.executeQuery(query);
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");

            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }

           
        } catch (Exception e) {
            if (this.statement != null) {
                this.statement.close();
                this.statement = null;
            }
            if (this.resultSet != null) {
                this.resultSet.close();
            }
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        }

    }

    /**
     * Schlie�t ein Statement das ResultSet (f�r Aufrufe mit executeQuery()
     * und get())
     * 
     * @throws java.lang.Exception
     * @see #executeQuery( String )
     * @see #get()
     */
    public void closeQuery() throws Exception {

        try {
            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            if (this.statement != null) {
                this.statement.close();
                this.statement = null;
            }
            if (this.resultSet != null) {
                this.resultSet.close();
            }

        } catch (Exception e) {
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        }

    }

    /**
     * F�hrt ein SQL-Statement aus
     * 
     * @param query
     * @return Anzahl betroffener S�tze f�r (INSERT, UPDATE oder DELETE)
     *         Statements, sonst 0
     * @throws java.lang.Exception
     *  
     */
    public int executeUpdate(String query) throws Exception {
        int rowCount = 0;
        Statement stmt = null;
        try {

            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + query);

            if (profiler != null) profiler.start(query);
            stmt = connection.createStatement();
            rowCount = stmt.executeUpdate(query);
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
            try {
                if (profiler != null) try {
                    profiler.stop("", "");
                } catch (Exception e) {
                }
            } catch (Exception e) {
            }

        } catch (Exception e) {
            try {
                if (profiler != null) try {
                    profiler.stop("ERROR", e.getMessage());
                } catch (Exception ex) {
                }
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            if (stmt != null) {
                stmt.close();
                stmt = null;
            }
            try {
                if (profiler != null) try {
                    profiler.stop("", "");
                } catch (Exception e) {
                }
            } catch (Exception e) {
            }
        }
        return rowCount;

    }

    /**
     * F�hrt ein Statement aus
     * 
     * @throws java.lang.Exception
     * @see #executeQuery( String )
     * @see #get()
     */
    public void execute(String query) throws Exception {

        Statement stmt = null;
        try {

            logger.debug9("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            query = normalizeStatement(query, replacement);
            logger.debug6(SOSClassUtil.getMethodName() + ": " + query);
            if (profiler != null) profiler.start(query);
            stmt = connection.createStatement();
            stmt.execute(query);
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
            if (profiler != null) try {
                profiler.stop("", "");
            } catch (Exception e) {
            }
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw new Exception(SOSClassUtil.getMethodName() + ":"
                    + e.toString(), e);
        } finally {
            if (stmt != null) {
                stmt.close();
                stmt = null;
            }
        }
    }

    /**
     * Transaction abschliessen
     * 
     * @throws Exception
     */
    public void commit() throws Exception {

        logger.debug9("calling " + SOSClassUtil.getMethodName());

        if (connection == null)
                throw new Exception(
                        SOSClassUtil.getMethodName()
                                + ": sorry, there is no successful connection established."
                                + " may be the connect method was not called");

        connection.commit();

        logger.debug6(SOSClassUtil.getMethodName() + " successfully executed.");

    }

    /**
     * Transaktion zurueckrollen
     * 
     * @throws Exception
     */
    public void rollback() throws Exception {

        logger.debug9("calling " + SOSClassUtil.getMethodName());

        if (connection == null)
                throw new Exception(
                        SOSClassUtil.getMethodName()
                                + ": sorry, there is no successful connection established."
                                + " may be the connect method was not called");

        connection.rollback();

        logger.debug6(SOSClassUtil.getMethodName() + " successfully executed.");

    }

    /**
     * AutoCommit-Modus setzen f�r automatischen Transaktionsabschluss.
     * 
     * @param autoCommit
     * @throws Exception
     */
    public void setAutoCommit(boolean autoCommit) throws Exception {

        if (connection == null)
                throw new Exception(
                        SOSClassUtil.getMethodName()
                                + ": sorry, there is no successful connection established."
                                + " may be the connect method was not called");
        connection.setAutoCommit(autoCommit);
    }

    /**
     * @return boolean den aktuellen AutoCommit-Status
     * @throws Exception
     */
    public boolean getAutoCommit() throws Exception {

        if (connection == null)
                throw new Exception(
                        SOSClassUtil.getMethodName()
                                + ": sorry, there is no successful connection established."
                                + " may be the connect method was not called");

        return connection.getAutoCommit();
    }

    /**
     * Liest den n�chsten Satz eines ResultSets. Das setzt vorraus, dass das
     * ResultSet mit {@link #executeQuery}erzeugt wurde.
     * 
     * @return HashMap on success, else null
     * @throws java.lang.Exception
     * @see #executeQuery
     * @see #execute( String )
     */
    public HashMap get() throws Exception {

        HashMap record = new LinkedHashMap();
        String columnName, columnValue;
        try {
            logger.debug6("Calling " + SOSClassUtil.getMethodName());
            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            if (profiler != null) try {
                profiler.start("");
            } catch (Exception e) {
            }
            if (this.resultSet.next()) {
                for (int i = 1; i <= this.resultSet.getMetaData()
                        .getColumnCount(); i++) {
                    columnName = this.resultSet.getMetaData().getColumnName(i);
                    columnValue = this.resultSet.getString(columnName);
                    if (SOSString.isEmpty(columnValue)) columnValue = "";
                    record.put(normalizeKey(columnName), columnValue.trim());
                }
            }
           
        } catch (Exception e) {
            if (profiler != null) try {
                profiler.stop("ERROR", e.getMessage());
            } catch (Exception ex) {
            }
            throw e;
        }
        if (profiler != null) try {
            profiler.stop("", "");
        } catch (Exception e) {
        }
        return record;
    }

    /**
     * Liefert die Anzahl Felder aus dem letzten ResultSet zur�ck.
     * 
     * @return int Anzahl Felder aus dem letzten ResultSet zur�ck.
     * @throws Exception
     * @see #execute( String )
     */
    public int fieldCount() throws Exception {

        if (connection == null)
                throw new Exception(
                        SOSClassUtil.getMethodName()
                                + ": sorry, there is no successful connection established."
                                + " may be the connect method was not called");

        return this.resultSet.getMetaData().getColumnCount();
    }

    /**
     * Liefert die Spaltennamen aus dem letzten ResultSet zur�ck.
     * 
     * @return String[] aus Spaltennamen
     * @throws java.lang.Exception
     * @see #execute( String )
     */
    public String[] fieldNames() throws Exception {

        int fieldCount = this.fieldCount();
        String[] fieldNames = new String[fieldCount];
        try {
            logger.debug6("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            for (int i = 0; i < fieldCount; i++) {
                fieldNames[i] = normalizeKey(this.resultSet.getMetaData()
                        .getColumnName(i + 1));
            }
            Arrays.sort(fieldNames);
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            throw e;
        }
        return fieldNames;
    }

    /**
     * Liefert den Feldnamen des eingegebenen Index eines ResultSets zur�ck.
     * 
     * @param index
     *            des Felds im ResultSet
     * @return String Feldname
     * @throws java.lang.Exception
     * @see #execute( String )
     */
    public String fieldName(int index) throws Exception {

        if (connection == null)
                throw new Exception(
                        SOSClassUtil.getMethodName()
                                + ": sorry, there is no successful connection established."
                                + " may be the connect method was not called");

        return this.resultSet.getMetaData().getColumnName(index);
    }

    /**
     * Liefert die Feldbeschreibung des eingegebenen Index eines ResultSets
     * zur�ck.
     * 
     * @param index
     *            des Felds im ResultSet
     * @return HashMap Feldbeschreibung
     * @throws java.lang.Exception
     * @see #execute( String )
     */
    public HashMap fieldDesc(int index) throws Exception {

        HashMap fieldDesc = new HashMap();
        try {

            logger.debug6("calling " + SOSClassUtil.getMethodName());

            if (connection == null)
                    throw new Exception(
                            SOSClassUtil.getMethodName()
                                    + ": sorry, there is no successful connection established."
                                    + " may be the connect method was not called");

            fieldDesc.put("columnDisplaySize", String.valueOf(this.resultSet
                    .getMetaData().getColumnDisplaySize(index)));
            fieldDesc.put("columnLabel", this.resultSet.getMetaData()
                    .getColumnLabel(index));
            fieldDesc.put("columnName", this.resultSet.getMetaData()
                    .getColumnName(index));
            fieldDesc.put("columnType", String.valueOf(this.resultSet
                    .getMetaData().getColumnType(index)));
            fieldDesc.put("columnTypeName", this.resultSet.getMetaData()
                    .getColumnTypeName(index));
            fieldDesc.put("precision", String.valueOf(this.resultSet
                    .getMetaData().getPrecision(index)));
            fieldDesc.put("scale", String.valueOf(this.resultSet.getMetaData()
                    .getScale(index)));
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            throw e;
        }
        return fieldDesc;
    }

    /**
     * Liefert ein normalisiertes SQL-Statement zur�ck. Die Meta-Funktionen:
     * %lcase, %ucase, %now, %updlock, %timestamp, %timestamp_iso werden in die
     * entsprechende SQL-Syntax umgesetzt.
     * 
     * @param inputString
     * @return String normalisiertes Statement
     * @throws java.lang.Exception
     */
    public String normalizeStatement(String inputString) throws Exception {

        return normalizeStatement(inputString, replacement);
    }

    /**
     * Liefert ein normalisiertes SQL-Statement zur�ck. Die Meta-Funktionen:
     * %lcase, %ucase, %now, %updlock, %timestamp, %timestamp_iso werden in die
     * entsprechende SQL-Syntax umgesetzt.
     * 
     * @param replacement
     * @param inputString
     * @return String normalisiertes Statement
     * @throws java.lang.Exception
     */
    public String normalizeStatement(String inputString, String[] replacement)
            throws Exception {

        logger.debug9("calling " + SOSClassUtil.getMethodName());

        logger.debug9(".. inputString= " + inputString);

        if (replacement.length < 4) return inputString;

        StringBuffer aPattern = new StringBuffer(
                "(%lcase|%ucase|%now|%updlock)|")
                .append(
                        "(%timestamp\\('[0-9]{1,2}\\.[0-9]{1,2}\\.[0-9]{4}'\\))|")
                .append(
                        "(%timestamp\\('[0-9]{1,2}\\.[0-9]{1,2}\\.[0-9]{4}[ \t]?[0-9]{2}:[0-9]{2}:[0-9]{2}'\\))|")
                .append(
                        "(%timestamp_iso\\('[0-9]{4}[-]?[0-9]{1,2}[-]?[0-9]{1,2}'\\))|")
                .append(
                        "(%timestamp_iso\\('[0-9]{4}[-]?[0-9]{1,2}[-]?[0-9]{1,2}[ \t]?[0-9]{2}:[0-9]{2}:[0-9]{2}'\\))");

        Pattern pattern = Pattern.compile(aPattern.toString());
        Matcher matcher = pattern.matcher(inputString);

        StringBuffer buffer = new StringBuffer();
        String replaceString = new String();

        boolean found = false;
        try {
            while ((found = matcher.find())) {
                replaceString = matcher.group();

                if (matcher.group(1) != null) {
                    replaceString = replaceString.replaceAll("%lcase",
                            replacement[0]);
                    replaceString = replaceString.replaceAll("%ucase",
                            replacement[1]);
                    replaceString = replaceString.replaceAll("%now",
                            replacement[2]);
                    replaceString = replaceString.replaceAll("%updlock",
                            replacement[3]);
                }

                if (matcher.group(2) != null || matcher.group(3) != null)
                        replaceString = this.toDate(this.toTimestamp(
                                replaceString, NLS_DE));

                if (matcher.group(4) != null || matcher.group(5) != null) {
                    replaceString = this.toDate(this.toTimestamp(replaceString,
                            NLS_ISO));
                }

                matcher.appendReplacement(buffer, replaceString);
            }
            matcher.appendTail(buffer);
            logger.debug9(SOSClassUtil.getMethodName()
                    + " successfully executed.");
        } catch (Exception e) {
            throw e;
        }
        return buffer.toString();
    }

    /**
     * stellt die Datenbankverbindung her.
     * 
     * @throws java.lang.Exception
     *             falls ein Datenbankfehler vorliegt.
     */
    abstract public void connect() throws Exception;

    /**
     * liefert timestamp Funktion zur�ck.
     * 
     * @param dateString
     * @return oracle timestamp Funktion
     * @throws java.lang.Exception
     */
    abstract public String toDate(String dateString) throws Exception;

    /**
     * schliesst alle offenen Resourcen.
     * 
     * @throws Exception
     */
    public void disconnect() throws Exception {

        if (this.statement != null) {
            this.statement.close();
            this.statement = null;
        }
        if (this.resultSet != null) {
            this.resultSet.close();
            this.resultSet = null;
        }
        if (connection != null) {
            connection.rollback();
            connection.close();
            connection = null;
            try {
                logger.debug6(SOSClassUtil.getMethodName()
                        + ": successfully disconnected.");
            } catch (Exception e) {
            }
        }
    }

    /**
     * Destruktor
     * 
     * @throws Throwable
     */
    protected void finalize() throws Throwable {

        try {
            this.disconnect();
        } catch (Exception e) {
        }
        super.finalize();
    }

    /**
     * Liefert einen Timestamp zur�ck.
     * 
     * @param dateString
     * @param nlsFormat
     * @return String Timestamp
     * @throws java.lang.Exception
     */
    public String toTimestamp(String dateString, String nlsFormat)
            throws Exception {
        Pattern p;
        String[] dateTime;
        String[] datePart = null;
        String[] timePart = { "00", "00", "00"};
        StringBuffer result = new StringBuffer();
        String datePatternSeparator;

        if (nlsFormat.equalsIgnoreCase(SOSConnection.NLS_DE))
            datePatternSeparator = "[.]";
        else if (nlsFormat.equalsIgnoreCase(SOSConnection.NLS_ISO)) {
            datePatternSeparator = "[-]";
        } else
            throw new Exception("invalid nlsFormat.");

        dateString = dateString.substring(dateString.indexOf(39) + 1,
                dateString.lastIndexOf(39));
        p = Pattern.compile("[ ]");

        dateTime = p.split(dateString);
        p = Pattern.compile(datePatternSeparator);
        datePart = p.split(dateTime[0]);
        if (dateTime.length > 1) {
            p = Pattern.compile("[:]");
            timePart = p.split(dateTime[1]);
        }
        if (nlsFormat.equalsIgnoreCase(SOSConnection.NLS_DE)) {
            result.append(datePart[2]).append("-").append(datePart[1]).append(
                    "-").append(datePart[0]);
        } else if (nlsFormat.equalsIgnoreCase(SOSConnection.NLS_ISO)) {
            result.append(datePart[0]).append("-").append(datePart[1]).append(
                    "-").append(datePart[2]);
        }
        result.append(" ").append(timePart[0]).append(":").append(timePart[1])
                .append(":").append(timePart[2]);

        return result.toString();
    }

    /**
     * Liefert den eingegebenen String in Kleinschreibung.
     * 
     * @param key
     * @return String
     * @throws java.lang.Exception
     * @see #setKeysToLowerCase
     * @see #setKeysToUpperCase
     */
    protected String normalizeKey(String key) throws Exception {

        if (SOSString.isEmpty(key))
                throw new Exception(getClass() + ":"
                        + SOSClassUtil.getMethodName() + ": invalid key.");
        if (this.lowerCase)
            return key.toLowerCase();
        else
            return key.toUpperCase();
    }

    /**
     * Aktiviert die Kleinschreibung f�r Feldnamen
     * 
     * @see #setKeysToUpperCase
     */
    public void setKeysToLowerCase() throws Exception {

        logger.debug6("calling " + SOSClassUtil.getMethodName());
        logger.debug9(".. now keys set to lower case.");
        lowerCase = true;
    }

    /**
     * Aktiviert die Grossschreibung f�r Feldnamen
     * 
     * @see #setKeysToLowerCase
     */
    public void setKeysToUpperCase() throws Exception {

        logger.debug6("calling " + SOSClassUtil.getMethodName());
        logger.debug9(".. now keys set to upper case.");
        lowerCase = false;
    }

    /**
     * Mehrere Statements aus einer Zeichenkette extrahieren. <br>
     * Statements m�ssen voneinander per ;\n getrennt sein Vorhandene ein- oder
     * mehrzeilige Kommentare werden aus den Statements entfernt
     * 
     * @param contentOfClobAttribute
     *            Zeichenkette mit den SQL Statements zum Extrahieren
     * @return statements ArrayList mit den einzelnen Statements
     * @throws Exception
     */
   
   public ArrayList getStatements(String contentOfClobAttribute)
            throws Exception {

        if (contentOfClobAttribute == null
                || contentOfClobAttribute.length() == 0) { throw new Exception(
                SOSClassUtil.getMethodName()
                        + ": contentOfClobAttribute is empty"); }

        logger.debug6(SOSClassUtil.getMethodName()
                + ": contentOfClobAttribute = " + contentOfClobAttribute);

        contentOfClobAttribute = contentOfClobAttribute
                .replaceAll("\r\n", "\n");

        //contentOfClobAttribute = contentOfClobAttribute.replaceAll(
        //        "[/][*](?s).*?[*][/]", "");

        ArrayList statements = new ArrayList();

        String splitString;
        String endString;

        boolean enableProcedureSearch = true;

        // a slash in a separate line concatenates ; to the preceeding line to
        // comply with procedural statements
        // Oracle ends PL/SQL statements with / in a separate line
        // MySQL ends procedural statements with \g in a separate line
        if (this instanceof SOSMySQLConnection) {
            splitString = "\n\\\\g\n";
            endString = "\n\\g\n";
            // endString = "";
        } else if (this instanceof SOSOracleConnection) {
            splitString = "\n/\n";
            //endString = "\n/\n";
            endString = "";
        } else {
            //splitString = ";";
            //endString = "";
            
            splitString = "\n/\n";
            endString = "";
            
            //enableProcedureSearch = false;

        }

        contentOfClobAttribute = this
                .stripOuterComments(contentOfClobAttribute);

        String[] commands = contentOfClobAttribute.replaceAll("\\;[ \\t]", ";")
                .split(splitString);

        for (int i = 0; i < commands.length; i++) {

            if (commands[i] == null || commands[i].trim().length() == 0) {
                continue;
            }

            String statement = commands[i].trim();

            if (enableProcedureSearch) {
                // eigentlich nicht richtig - kann END [Prozedurenname]
                // vorkommen
                // Es handelt sich um ein PL/SQL Block
                if (statement.toLowerCase().endsWith("end")
                        || statement.toLowerCase().endsWith("end;")) {
                    //Pr�fen ob mit PL/SQL anf�ngt
                    if (this.isProcedureSyntax(statement)) {
                        // PL/SQL Block
                        statements.add(statement + endString);
                            logger.debug6(SOSClassUtil.getMethodName()
                                + " : statement =" + statement + endString);
                    } else {
                        //kein PL/SQL Anfang
                        String s = statement.replaceAll(this.replaceBackslash, this.replacementBackslash);
                        s = s.replaceAll(this.replaceDoubleApostrophe, this.replacementDoubleApostrophe);
                        
                        this.splitStatements(statements, new StringBuffer(s.trim()), null,
                               endString, true,0);
                        
                        
                        if (!this.beginProcedure.equals("")) {
                            int posPLSQL = statement.indexOf(this.beginProcedure);
                            String substatement = statement.substring(posPLSQL);
                       
                            statements.add(substatement + endString);
                                logger.debug6(SOSClassUtil.getMethodName()
                                        + " : statement =" + substatement
                                        + endString);
                   
                        }
                    }

                } // if endWith(END)
                else {
                    // entweder SQL oder Pl/SQL zB drop procedure
                    String s = statement.replaceAll(this.replaceBackslash, this.replacementBackslash);
                        s = s.replaceAll(this.replaceDoubleApostrophe,this.replacementDoubleApostrophe);
                        
                    this.splitStatements(statements, new StringBuffer(s.trim()), null,
                                 endString, false,0);
                    
                }
            }// enable PL/SQLsearch
            else {
                String s = statement.replaceAll(this.replaceBackslash, this.replacementBackslash);
                s = s.replaceAll(this.replaceDoubleApostrophe, this.replacementDoubleApostrophe);
                
                this.splitStatements(statements, new StringBuffer(s.trim()), null,
                         endString, false,0);
           }
        }

        return statements;
    }
    
     /**
     * @param statements
     * @param stmt
     * @param position
     * @param procedurEnd
     * @param returnProcedureBegin
     * @throws Exception
     */
    private void splitStatements(ArrayList statements, StringBuffer st,
            Integer position, String procedurEnd, boolean returnProcedureBegin,int count)
            throws Exception {
        
        //System.gc();
   
        //System.out.println("count : "+count);
        
        this.beginProcedure = "";
              
        count += 1;

        StringBuffer sub;

        int semicolon = -1;
        int apostrophe_first = -1;

        if (position == null) {
            semicolon = st.indexOf(";");
            apostrophe_first = st.indexOf("'");
        } else {
            semicolon = st.indexOf(";", position.intValue());
            apostrophe_first = st.indexOf("'", position.intValue());
        }
         
        // Hochkommas hinter dem ; oder gar nicht gefunden
        if ((apostrophe_first > semicolon) || apostrophe_first == -1) {

            String value = "";
            
            if(semicolon == -1){
                value = st.toString().trim();
            }
            else{
                value = st.toString().substring(0, semicolon).trim();
            }
             
            value = value.replaceAll(this.replacementBackslash, this.replaceBackslash);
            value = value.replaceAll(this.replacementDoubleApostrophe,
                    this.replaceDoubleApostrophe);
            
            if (this.isProcedureSyntax(value)) {
                
               
                if (returnProcedureBegin) {
                    this.beginProcedure = value;
                    return;
                } else {
                    // hier nur die einfache Proceduren
                    // create, drop usw
                    if (!procedurEnd.equals("")) {
                            value += procedurEnd;
                    }
                }
            }
            
            if(!value.equals("")){
                statements.add(value);
                    logger.debug6(SOSClassUtil.getMethodName()
                            + ": statement = " + value);
            }

            if(semicolon != -1){
                sub = new StringBuffer(st.substring(semicolon + 1));
             
                if (sub != null && sub.length() != 0) {
                    this.splitStatements(statements, sub, null, procedurEnd,
                            returnProcedureBegin,count);
                }
            }
            
        } 
        else {
            // Hochkommas gefunden
            int apostrophe_second = st.indexOf("'", apostrophe_first + 1);
            // egal ob Hochkomma vor ; oder danach
            if(apostrophe_second != -1){
                this.splitStatements(statements, st, new Integer(
                    apostrophe_second + 1), procedurEnd, returnProcedureBegin,count);
            }
            else{
                throw new Exception("Schliessende Hochkomma nicht gefunden !!!!!!!!!!!!!!!! = "+apostrophe_first+" = "+st);
            }
        }
    }

      /**
     * @param statement
     * @return @throws
     *         Exception
     */
    private boolean isProcedureSyntax(String statement) throws Exception {

        if (statement == null) { throw new Exception("statement is empty"); }

        statement = statement.toLowerCase().trim();

        if (statement.startsWith("procedure")
                || statement.startsWith("function")
                || statement.startsWith("declare")
                || statement.startsWith("begin")) { return true; }

        StringBuffer patterns = new StringBuffer();
        patterns.append("^create+[\\s]*procedure");
        patterns.append("|^create+[\\s]*function");
        patterns.append("|^create+[\\s]*package");
        patterns.append("|^create+[\\s]*trigger");
        patterns.append("|^drop+[\\s]*procedure");
        patterns.append("|^drop+[\\s]*function");
        patterns.append("|^drop+[\\s]*package");
        patterns.append("|^drop+[\\s]*trigger");
        patterns.append("|^create+[\\s]*or+[\\s]*replace+[\\s]*procedure");
        patterns.append("|^create+[\\s]*or+[\\s]*replace+[\\s]*function");
        patterns.append("|^create+[\\s]*or+[\\s]*replace+[\\s]*package");
				patterns.append("|^create+[\\s]*or+[\\s]*replace+[\\s]*trigger");

        Pattern p = Pattern.compile(patterns.toString());
        Matcher matcher = p.matcher(statement);
        if (matcher.find()) { return true; }

        return false;
    }

    
    /**
     * Kommentare , die zu keinem Statement gehren, entfernen
     * zB einzeilige #,-- ,//
     * oder mehzeilige - die Zeile mu mit /* anfangen und beim schliessen eine neue Zeile fr *
     *  
     * @param statement
     * @return
     * @throws Exception
     */
    private String stripOuterComments(String statement) throws Exception {
        if (statement == null) { throw new Exception("statement is empty"); }

        StringBuffer sb = new StringBuffer("");
        
        StringTokenizer st = new StringTokenizer(statement,"\n");
    
        boolean addRow = true;
        
        while (st.hasMoreTokens()) {
            String row = st.nextToken().trim();
            if (row == null || row.length() == 0) continue;

            if (row.startsWith("--") || row.startsWith("//") || row.startsWith("#")) {
                continue;
            }
            // wir entfernen nur die Kommentare, die am Zeilebeginn stehen()
            row = row.replaceAll("^[/][*](?s).*?[*][/][\\s]*;*",""); 
            
            // Das machen wir nicht , weil kann in value vorkommen
            //if(row.indexOf("//") != -1) {
            //    row = row.substring(0, row.indexOf("//"));
            //}
        
            if (row.length() == 0) continue;

            if(row.startsWith("/*")){
                addRow = false;
                continue;
            }
            //Verschachtelter Syntax wird bei der Kommentaren  nicht unterschttzt, 
            //daswegen ruhig endWith, sonst knnte */ innerhalb von /* */ vorkommen
            if(row.endsWith("*/") && addRow == false){
                addRow = true;
                continue;
            }
            
            if(addRow == false){
             continue;    
            }
            
            sb.append(row + "\n");
        }

        return sb.toString();
    }


    /**
     * Mehrere Statements aus einer Zeichenkette extrahieren und ausf�hren.
     * <br>
     * 
     * @param contentOfClobAttribute
     *            Zeichenkette mit den SQL Statements zum Extrahieren
     * @throws Exception
     */
    public void executeStatements(String contentOfClobAttribute)
            throws Exception {

        ArrayList statements = null;

        logger.debug6(SOSClassUtil.getMethodName()
                + " : contentOfClobAttribute = " + contentOfClobAttribute);

        try {
            statements = this.getStatements(contentOfClobAttribute);

            if (statements == null || statements.size() == 0) { throw new Exception(
                    "no sql statements found"); }

            boolean hasOpenResultSet = false;

            for (int i = 0; i < statements.size(); i++) {

                String statement = statements.get(i).toString().trim();

                if (statement.toLowerCase().startsWith("select")) {

                    if (hasOpenResultSet) this.closeQuery();

                    this.executeQuery(statement);

                    hasOpenResultSet = true;
                } else {
                    this.executeUpdate(statement);
                }
            }

        } catch (Exception e) {
            throw new Exception(SOSClassUtil.getMethodName() + " : " + e, e);
        }
    }

    /**
     * Zuweisen des Profilers. Wenn profiler_ null ist, dann bedeutet das, das
     * Profiler nicht eingeschaltet ist. Ist die profiler_ nicht null, so ist
     * der Profiler eingeschaltet.
     * 
     * @param profiler_
     */
    public void setProfiler(SOSProfiler profiler_) {
        profiler = profiler_;
    }

    /**
     * setzt die Schreibweise der Feldname auf Grossschreibung.
     * 
     * @param value
     *            falls true beleibt bei default (Grossschreibung)
     */
    public void setFieldNameToUpperCase(boolean value) {
        fieldNameUpperCase = value;
    }

    /**
     * setzt die Schreibweise der Tabellenname auf Grossschreibung.
     * 
     * @param value
     *            falls true beleibt bei default (Grossschreibung)
     */
    public void setTableNameToUpperCase(boolean value) {
        tableNameUpperCase = value;
    }

    /**
     * liefert das Result Set einer mit execute() ausgef�hrten Abfrage
     * 
     * @return Returns the resultSet.
     */
    public ResultSet getResultSet() {
        return resultSet;
    }

    /**
     * @return Returns the url.
     */
    public String getUrl() {
        return url;
    }

		/**
     * liefert weitere Ausgaben vom Datenbankserver (nicht das Resultset)
     * TODO: auch fr MySQL und MSSQL     
     * @return Vector aus Strings fr jede Zeile des Outputs
     */
	  public Vector getOutput() throws Exception{
		  Vector out = new Vector();
		  return out;
	  }
}